import json
import heapq
import redis
from .json_web_token import JsonWebToken
from .auth_failed import AuthFailed
from django.conf import settings

JWT_CONF = getattr(settings, 'JWT_CONF', {})
MAX_LOGIN = JWT_CONF.get('max_login', 1)
REDIS_POOL = JWT_CONF.get('redis_pool') or redis.ConnectionPool(host='127.0.0.1', port=6379, db=0)
rds = redis.Redis(connection_pool=REDIS_POOL)


class JsonWebTokenFuse(JsonWebToken):

    def token(self, payload: dict):
        """根据payload生成token，
        :param payload:
            必选参数有: 1.exp 过期时间 ，now > exp 则失效
                      2.iat：签发时间戳 。
                      3.uid：用于辨识用户的唯一id，建议[uid, user]任选一种作为参数名都可以
            可选参数有：
                      1.iss:签发人，在同一台服务器中如果存在多个项目（多项目共用了Redis数据池），
                      如果这些项目不是共享token，那么在JwtFuse().auth()时为了防止token串号，可以设置iss用于区分。
        :return: 生成的token"""

        if 'iat' not in payload:
            raise ValueError('payload中不存在iat(签发时间)！')

        return super().token(payload=payload)

    def _generate_key(self, payload: dict):
        """生成token在Redis中的键名"""
        uid = payload.get('uid') or payload.get('user')  # 用户表示在payload中为user或uid
        if not uid:
            raise TypeError('payload中没有配置uid或user')
        iss = payload.get('iss') or 'no_iss'  # 如果没有指定签发人 则把iss默认为'no_iss'
        token_key = 'token:' + iss + ':' + str(uid)  # token在Redis中的键名
        return token_key

    def _push_token(self, tokens: list, token: str) -> list:
        """
        将token push到tokens中，如果tokens的长度超出max_len会pop掉时间最久的那条token
        :param tokens:一个保存token的list
        :param token:push的token，不一定保证push成功，这是因为如果该条token时间久远，在存入后即会被pop掉
        :return:tokens
        """
        _, payload, *_ = token.split('.')
        payload = json.loads(self.enc.debase64(payload, alt_chars='-_'))
        token_item = (payload['iat'], token)  # iat签发时间

        if token_item not in tokens:
            heapq.heappush(tokens, token_item)

        if len(tokens) > MAX_LOGIN:
            heapq.heappop(tokens)

        return tokens

    def auth(self, token: str):
        """验证token 只允许max_len条token有效"""
        payload = super().auth(token)
        if not payload:  # 如果token没有验证通过，返回payload（此时payload为AuthFailed()对象）
            return payload

        token_key = self._generate_key(payload)  # 生成token在Redis中的键名

        tokens = rds.get(token_key)  # 从redis中取出来的是bytes
        tokens = json.loads(tokens.decode('utf-8')) if tokens else []  # decode后转化成列表
        tokens = [tuple(i) for i in tokens]  # [[time,token],]-> [(time,token),],因为json后元组变成了列表
        tokens = self._push_token(tokens, token)  # token push to tokens
        # 存入redis的过期时间为3600秒 意味着3600秒后该token可以正常访问
        rds.set(token_key, json.dumps(tokens), 3600)
        if token in [i[-1] for i in tokens]:
            return payload
        else:
            return AuthFailed('token超出允许的数量范围')  # 103: token超出允许的数量范围



import re


def judge_pc_or_mobile(ua):
    """
    判断访问来源是pc端还是手机端
    :param ua: 访问来源头信息中的User-Agent字段内容
    :return:
    """
    factor = ua
    is_mobile = False
    _long_matches = r'googlebot-mobile|android|avantgo|blackberry|blazer|elaine|hiptop|ip(hone|od)|kindle|midp|mmp' \
                    r'|mobile|o2|opera mini|palm( os)?|pda|plucker|pocket|psp|smartphone|symbian|treo|up\.(browser|link)' \
                    r'|vodafone|wap|windows ce; (iemobile|ppc)|xiino|maemo|fennec'
    _long_matches = re.compile(_long_matches, re.IGNORECASE)
    _short_matches = r'1207|6310|6590|3gso|4thp|50[1-6]i|770s|802s|a wa|abac|ac(er|oo|s\-)|ai(ko|rn)|al(av|ca|co)' \
                     r'|amoi|an(ex|ny|yw)|aptu|ar(ch|go)|as(te|us)|attw|au(di|\-m|r |s )|avan|be(ck|ll|nq)|bi(lb|rd)' \
                     r'|bl(ac|az)|br(e|v)w|bumb|bw\-(n|u)|c55\/|capi|ccwa|cdm\-|cell|chtm|cldc|cmd\-|co(mp|nd)|craw' \
                     r'|da(it|ll|ng)|dbte|dc\-s|devi|dica|dmob|do(c|p)o|ds(12|\-d)|el(49|ai)|em(l2|ul)|er(ic|k0)|esl8' \
                     r'|ez([4-7]0|os|wa|ze)|fetc|fly(\-|_)|g1 u|g560|gene|gf\-5|g\-mo|go(\.w|od)|gr(ad|un)|haie|hcit' \
                     r'|hd\-(m|p|t)|hei\-|hi(pt|ta)|hp( i|ip)|hs\-c|ht(c(\-| |_|a|g|p|s|t)|tp)|hu(aw|tc)|i\-(20|go|ma)' \
                     r'|i230|iac( |\-|\/)|ibro|idea|ig01|ikom|im1k|inno|ipaq|iris|ja(t|v)a|jbro|jemu|jigs|kddi|keji' \
                     r'|kgt( |\/)|klon|kpt |kwc\-|kyo(c|k)|le(no|xi)|lg( g|\/(k|l|u)|50|54|e\-|e\/|\-[a-w])|libw|lynx' \
                     r'|m1\-w|m3ga|m50\/|ma(te|ui|xo)|mc(01|21|ca)|m\-cr|me(di|rc|ri)|mi(o8|oa|ts)|mmef|mo(01|02|bi' \
                     r'|de|do|t(\-| |o|v)|zz)|mt(50|p1|v )|mwbp|mywa|n10[0-2]|n20[2-3]|n30(0|2)|n50(0|2|5)|n7(0(0|1)' \
                     r'|10)|ne((c|m)\-|on|tf|wf|wg|wt)|nok(6|i)|nzph|o2im|op(ti|wv)|oran|owg1|p800|pan(a|d|t)|pdxg' \
                     r'|pg(13|\-([1-8]|c))|phil|pire|pl(ay|uc)|pn\-2|po(ck|rt|se)|prox|psio|pt\-g|qa\-a|qc(07|12|21' \
                     r'|32|60|\-[2-7]|i\-)|qtek|r380|r600|raks|rim9|ro(ve|zo)|s55\/|sa(ge|ma|mm|ms|ny|va)|sc(01|h\-' \
                     r'|oo|p\-)|sdk\/|se(c(\-|0|1)|47|mc|nd|ri)|sgh\-|shar|sie(\-|m)|sk\-0|sl(45|id)|sm(al|ar|b3|it' \
                     r'|t5)|so(ft|ny)|sp(01|h\-|v\-|v )|sy(01|mb)|t2(18|50)|t6(00|10|18)|ta(gt|lk)|tcl\-|tdg\-|tel(i|m)' \
                     r'|tim\-|t\-mo|to(pl|sh)|ts(70|m\-|m3|m5)|tx\-9|up(\.b|g1|si)|utst|v400|v750|veri|vi(rg|te)' \
                     r'|vk(40|5[0-3]|\-v)|vm40|voda|vulc|vx(52|53|60|61|70|80|81|83|85|98)|w3c(\-| )|webc|whit' \
                     r'|wi(g |nc|nw)|wmlb|wonu|x700|xda(\-|2|g)|yas\-|your|zeto|zte\-'

    _short_matches = re.compile(_short_matches, re.IGNORECASE)
    if _long_matches.search(factor) is not None:
        is_mobile = True
    user_agent = factor[0:4]
    if _short_matches.search(user_agent) is not None:
        is_mobile = True

    return is_mobile



import hmac
import base64


class Encryption:
    """加密类"""

    def to_bytes(self, bytes_or_str):
        if isinstance(bytes_or_str, str):
            value = bytes_or_str.encode('utf-8')
        else:
            value = bytes_or_str
        return value

    def to_str(self, bytes_or_str):
        if isinstance(bytes_or_str, bytes):
            value = bytes_or_str.decode('utf-8')
        else:
            value = bytes_or_str
        return value

    def base64(self, content, alt_chars=None):
        """base64编码
        :param content:需要编码的内容
        为了生成安全的Base64字符串会把结尾的“==”号去掉。
        :param alt_chars:可选的alt_chars是长度为2的字节字符串（推荐：alt_chars="-_"），
        指定“+”和“/”字符的替换字符，用于生成安全的Base64字符串“+” 替换为 “-”，“/” 替换为 ""_"" ,也可以为None
        :return:base64编码的数据，一般为字符串类型
        """
        content = self.to_bytes(content)
        alt_chars = self.to_bytes(alt_chars) if alt_chars else alt_chars
        content_base64 = base64.b64encode(content, altchars=alt_chars)
        return content_base64.decode('utf-8').replace('=', '')

    def debase64(self, content: str, is_string=True, alt_chars=None):
        """base64解码
        :param content:需要解码的内容，一般为字符串类型，
        base64数据长度应为4的整数倍，如果长度取模后的余数不等于0，那么需要补齐的“=”号的个数为4减去取模后的余数。
        :param is_string:解码结果的期待数据类型是否为字符串，is_string=True时编码结果会decode()成字符串。
        如果是图片这里需要is_string=False以返回bytes
        :param alt_chars:可选的alt_chars是长度为2的字节字符串（一般为：alt_chars="-_"），
        需要解码的内容如果是安全的Base64字符串,需要指定“+”和“/”字符的替换字符。
        :return:ase64解码的数据
        """
        num = len(content) % 4  # content长度取模后的余数
        content = content + '=' * (4 - num) if num else content  # 补齐“=”号，个数为4减去余数
        content = self.to_bytes(content)
        alt_chars = self.to_bytes(alt_chars) if alt_chars else alt_chars
        content = base64.b64decode(content, altchars=alt_chars)
        return content.decode('utf-8') if is_string else content

    def hmac_sha(self, key, message, digest_mod='sha256'):
        """
        hmac.SHA 加密
        :param key: 密钥
        :param message:加密内容 bytes
        :param digest_mod:加密类型 sha256 或者 sha1等
        :return:密文 base64格式的密文
        """
        key = self.to_bytes(key)
        message = self.to_bytes(message)
        h = hmac.new(key, message, digestmod=digest_mod)
        return self.base64(h.digest(), alt_chars="-_")


if __name__ == '__main__':
    print(Encryption().hmac_sha('secret', 'Hello world'))



from distutils.core import setup

# from os import path as os_path
#
# this_directory = os_path.abspath(os_path.dirname(__file__))
#
#
# def read_file(filename):
#     with open(os_path.join(this_directory, filename)) as f:
#         long_description = f.read()
#         return long_description


setup(
    name="token-fuse",
    version="1.2.3",
    author="rui",
    packages=['token_fuse'],
    author_email="604729765@qq.com",
    url="https://www.leetab.com",
    description="json web token的生成和验证",
    install_requires=["redis>=3.5.3"],
    long_description=open("README.rst", encoding="utf-8").read(),
)
# python setup.py sdist
# twine upload dist/*



import json
import time

from .encryption import Encryption
from .auth_failed import AuthFailed
from django.conf import settings

SECRET_KEY = settings.SECRET_KEY


class JsonWebToken:
    def __init__(self):
        self.__header = json.dumps({"alg": "HS256", "typ": "JWT"})
        self.enc = Encryption()  # 加密类

    def _signature(self, header, payload):
        """注意这里的header和payload是base64编码之后的内容;"""
        signature = header + '.' + payload
        signature = self.enc.hmac_sha(SECRET_KEY, signature, digest_mod='sha256')
        return signature

    def token(self, payload: dict):
        """根据payload生成token，
        :param payload:
            必选参数有: 1.exp 过期时间 ，now > exp 则失效
                      2.uid：用户辨识用户的唯一id，建议[uid, user]任选一种作为参数名都可以
            可选参数有: 1.iss:签发人，
                      2.iat：签发时间戳 。
        :return: 生成的token
        """
        if 'exp' not in payload:
            raise ValueError('payload中不存在exp!')

        if 'uid' not in payload and 'user' not in payload:
            raise ValueError('payload中不存在uid(或user)！')

        payload = json.dumps(payload)  # 将payload从dict转为str
        header = self.enc.base64(self.__header, alt_chars='-_')  # base64加密header
        payload = self.enc.base64(payload, alt_chars='-_')  # 计算签名后才能将payload转换成base64
        signature = self._signature(header, payload)  # 计算签名
        token = header + '.' + payload + '.' + signature
        return token

    def auth(self, token: str):
        """
        验证token的合法性,这里只是验证token是否合法以及是否过期；
        :param token:待验证的token,exp(到期时间戳)是payload中的必选参数
        :return:token如果成功返回payload, 否则返回AuthFailed()的实例
        """
        try:
            header, payload, signature, *_ = token.split('.')
        except ValueError:
            return AuthFailed('token格式错误！应为：Header.Payload.Signature.')

        auth_sign = self._signature(header, payload)  # 用token中的header和payload生成签名
        if not auth_sign == signature:  # 如果token中header,payload与签名不符说明被篡改
            return AuthFailed('signature校验不通过，token不合法或者被篡改！')

        try:
            payload = json.loads(self.enc.debase64(payload, alt_chars='-_'))
        except Exception:
            return AuthFailed('payload部分格式错误！应该是一个被base64加密后的对象.')

        exp = payload['exp']  # 过期时间
        if int(time.time()) > exp:  # 30d=2592000s 90d=7776000s 是否到期
            return AuthFailed('token已经过期')  # 102: token已经过期
        return payload



class AuthFailed:
    """用于生成一个包含具体错误信息的布尔值为假的对象
    应用场景：当检查的token不合法时如果仅仅返回False，将无法了解该token的具体错误信息是什么，
    是被篡改还是已过期还是其它错误类型。AuthFailed(err_msg)生成的对象布尔值为False，同时包含了具体了错误信息"""

    def __init__(self, err_msg):
        """
        :param err_msg: 错误信息
        """
        self.__err_msg = err_msg

    def __bool__(self):
        return False

    def __str__(self):
        return self.__err_msg
